from mkdocs_puml.theme import Theme


def test_theme(diagram_and_encoded: tuple[str, str]):
    url = "example.url"
    name = "default/light"
    theme = Theme(url=url)
    with_theme = theme.include(name, diagram_and_encoded[0]).split("\n")

    assert with_theme[1] == f"!include {url}/{name}.puml"


def test_without_startuml():
    diagram = "Bob -> Alice"
    url = "example.url"
    name = "default/light"
    theme = Theme(url=url)

    with_theme = theme.include(name, diagram).split("\n")

    assert with_theme[0] == f"!include {url}/{name}.puml"


def test_with_c4(c4_diagram: str):
    url = "example.url"
    name = "default/light"
    theme = Theme(url=url)

    with_theme = theme.include(name, c4_diagram).split("\n")

    # strip because c4_diagram uses raw-formatted strings
    assert with_theme[1].strip() == "!include https://raw.git.../C4-PlantUML/master/C4_Container.puml"
    assert with_theme[2].strip() == f"!include {url}/{name}.puml"



from pathlib import Path
from unittest.mock import MagicMock

import pytest
from pytest_httpx import HTTPXMock


BASE_PUML_URL = "http://mocked/"
BASE_PUML_KEYWORD = "puml"
CUSTOM_PUML_KEYWORD = "plantuml"
TESTDATA_DIR = Path(__file__).resolve().parent.joinpath("testdata")


@pytest.fixture(scope="package")
def diagram_and_encoded():
    """The fixture to return puml diagram and
    the encoded by plantuml.com string
    """
    return (
        "@startuml\nBob -> Alice : hello\n@enduml",
        "SoWkIImgAStDuNBAJrBGjLDmpCbCJbMmKiX8pSd9vt98pKi1IW80",
    )


@pytest.fixture(scope="package")
def c4_diagram():
    return """
    @startuml
    !include https://raw.git.../C4-PlantUML/master/C4_Container.puml
    System(test, "Test", "Test system")
    @enduml
    """


@pytest.fixture(scope="package")
def svg_diagram():
    with open(TESTDATA_DIR.joinpath("plantuml.svg")) as f:
        return f.read()


@pytest.fixture
def mock_requests(httpx_mock: HTTPXMock, svg_diagram):
    def expect(call_count: int):
        for _ in range(call_count):
            httpx_mock.add_response(content=svg_diagram.encode("utf-8"))

    return expect


@pytest.fixture
def mock_requests_fallback(httpx_mock: HTTPXMock):
    def expect(call_count: int):
        for _ in range(call_count):
            httpx_mock.add_response(status_code=509, content=b"error")

    return expect


@pytest.fixture
def patch_path_mkdir(monkeypatch):
    """Patches pathlib.Path to return a MagicMock,
    preventing direct file system interactions
    """
    mock = MagicMock()
    monkeypatch.setattr("pathlib.Path.mkdir", mock)



import uuid
from unittest.mock import MagicMock

import jinja2
import pytest

from uuid import UUID

from mkdocs_puml.config import CacheConfig, InteractionConfig, LocalCacheConfig, PlantUMLConfig, ThemeConfig
from mkdocs_puml.model import ThemeMode
from mkdocs_puml.plugin import Diagram, PlantUMLPlugin
from tests.conftest import BASE_PUML_URL, TESTDATA_DIR


def is_uuid_valid(uuid_str: str) -> bool:
    """Check whether uuid_str is valid uuid4

    Args:
        uuid_str: uuid to test

    Returns:
        Boolean designating if uuid_str is valid
    """
    try:
        uuid_obj = UUID(uuid_str, version=4)
    except ValueError:
        return False
    return str(uuid_obj) == uuid_str


def patch_plugin_to_single_theme(plugin: PlantUMLPlugin):
    plugin.config.theme.enabled = False
    plugin.themer = None
    plugin.theme_light = None
    plugin.theme_dark = None


@pytest.fixture
def plugin_config() -> PlantUMLConfig:
    c = PlantUMLConfig()

    t = ThemeConfig()
    t.load_dict({"light": "default/light", "dark": "default/dark", "url": "test.url/themes"})

    cache = CacheConfig()
    cache.load_dict({"backend": "disabled", "local": LocalCacheConfig()})

    inter = InteractionConfig()
    inter.load_dict({"enabled": True})
    c.load_dict(
        {
            "puml_url": BASE_PUML_URL,
            "extra_css": [],
            "extra_javascript": [],
            "theme": t,
            "cache": cache,
            "interaction": inter
        }
    )
    return c


@pytest.fixture
def plugin_environment():
    loader = jinja2.FileSystemLoader(searchpath=TESTDATA_DIR)
    return jinja2.Environment(loader=loader)


@pytest.fixture
def plant_uml_plugin(plugin_config):
    plugin = PlantUMLPlugin()
    plugin.config = plugin_config
    plugin.on_config(plugin_config)

    return plugin


@pytest.fixture
def diagrams_dict(diagram_and_encoded):
    return {
        str(uuid.uuid4()): Diagram(diagram_and_encoded[0], mode=ThemeMode.LIGHT),
        str(uuid.uuid4()): Diagram(diagram_and_encoded[0], mode=ThemeMode.DARK),
        str(uuid.uuid4()): Diagram(diagram_and_encoded[0], mode=ThemeMode.LIGHT),
    }


@pytest.fixture(scope="package")
def md_lines():
    with open(TESTDATA_DIR.joinpath("markdown.md")) as f:
        return f.readlines()


@pytest.fixture
def html_page(plugin_environment, diagrams_dict):
    page = MagicMock(title="Test, page", file=MagicMock(), config=MagicMock())
    template = plugin_environment.get_template("output.html")
    page.content = template.render(
        uuid_class=PlantUMLPlugin.pre_class_name, uuids=diagrams_dict.keys()
    )
    page.html = page.content
    return page



import os

import pytest
from mkdocs_puml.model import Count
from mkdocs_puml.storage import FileStorage, RAMStorage
from mkdocs_puml.plugin import PlantUMLPlugin, ThemeMode
from mkdocs_puml.puml import Fallback, PlantUML
from mkdocs_puml.theme import Theme
from tests.conftest import BASE_PUML_KEYWORD, CUSTOM_PUML_KEYWORD
from tests.plugins.conftest import is_uuid_valid, patch_plugin_to_single_theme


def test_on_config(plugin_config):
    plugin = PlantUMLPlugin()
    plugin.config = plugin_config

    plugin.on_config(plugin_config)

    assert isinstance(plugin.puml, PlantUML)
    assert isinstance(plugin.themer, Theme)
    assert isinstance(plugin.storage, RAMStorage)

    assert plugin.theme_light == "default/light"
    assert plugin.theme_dark == "default/dark"
    assert plugin.puml_keyword == BASE_PUML_KEYWORD
    assert "assets/mkdocs_puml/puml.css" in plugin_config["extra_css"]
    assert "assets/mkdocs_puml/puml.js" in plugin_config["extra_javascript"]

    assert "assets/mkdocs_puml/interaction.css" in plugin_config["extra_css"]
    assert "assets/mkdocs_puml/interaction.js" in plugin_config["extra_javascript"]


def test_on_config_theme_disabled(plugin_config):
    plugin = PlantUMLPlugin()
    plugin_config.theme.enabled = False
    plugin.config = plugin_config

    plugin.on_config(plugin_config)

    assert isinstance(plugin.puml, PlantUML)
    assert plugin.themer is None

    assert plugin.theme_light is None
    assert plugin.theme_dark is None
    assert "assets/mkdocs_puml/puml.css" in plugin_config["extra_css"]
    assert "assets/mkdocs_puml/puml.js" in plugin_config["extra_javascript"]


def test_on_config_file_storage(plugin_config, patch_path_mkdir):
    plugin = PlantUMLPlugin()
    plugin.config = plugin_config
    plugin_config.cache.backend = "local"
    plugin_config.cache.local.load_dict({"path": "test"})

    plugin.on_config(plugin_config)

    assert isinstance(plugin.storage, FileStorage)


def test_on_config_interaction_disabled(plugin_config):
    plugin_config.interaction.enabled = False

    plugin = PlantUMLPlugin()
    plugin.config = plugin_config

    plugin.on_config(plugin_config)

    assert isinstance(plugin.puml, PlantUML)
    assert isinstance(plugin.themer, Theme)
    assert isinstance(plugin.storage, RAMStorage)

    assert plugin.theme_light == "default/light"
    assert plugin.theme_dark == "default/dark"
    assert plugin.puml_keyword == BASE_PUML_KEYWORD
    assert "assets/mkdocs_puml/puml.css" in plugin_config["extra_css"]
    assert "assets/mkdocs_puml/puml.js" in plugin_config["extra_javascript"]

    assert "assets/mkdocs_puml/interaction.css" not in plugin_config["extra_css"]
    assert "assets/mkdocs_puml/interaction.js" not in plugin_config["extra_javascript"]


def test_on_page_markdown_single_theme(plant_uml_plugin, md_lines):
    patch_plugin_to_single_theme(plant_uml_plugin)

    plant_uml_plugin.on_page_markdown("\n".join(md_lines))

    assert len(plant_uml_plugin.storage.items()) == 2

    for key in plant_uml_plugin.storage.keys():
        assert is_uuid_valid(key)

    for val in plant_uml_plugin.storage.schemes().values():
        assert "@startuml" in val and "@enduml" in val


def test_on_page_markdown_dual_themes(plant_uml_plugin, md_lines):
    plant_uml_plugin.on_page_markdown("\n".join(md_lines))

    assert len(plant_uml_plugin.storage.keys()) == 4

    for key, val in plant_uml_plugin.storage.items():
        if val.mode == ThemeMode.LIGHT:
            assert is_uuid_valid(key)
        else:
            uuid_key, _, dark = key.rpartition("-")
            assert is_uuid_valid(uuid_key)
            assert dark == "dark"

    for val in plant_uml_plugin.storage.schemes().values():
        assert "@startuml" in val and "@enduml" in val


def test_on_page_markdown_custom_keyword(plant_uml_plugin, md_lines):
    # Test if PlantUML diagrams are correctly extracted with a custom keyword
    plant_uml_plugin.config.puml_keyword = CUSTOM_PUML_KEYWORD
    plant_uml_plugin.on_page_markdown("\n".join(md_lines))

    assert (
        len(plant_uml_plugin.storage.items()) == 4
    )  # 2 (light / dark) on each diagram


def test_on_env(mock_requests, plant_uml_plugin, diagrams_dict, plugin_environment):
    mock_requests(len(diagrams_dict))

    plant_uml_plugin.storage.data = diagrams_dict
    plant_uml_plugin.on_env(plugin_environment)

    for _, diagram in plant_uml_plugin.storage.items():
        assert diagram.diagram.startswith("<svg")


def test_on_env_fallback(
    mock_requests_fallback, plant_uml_plugin, diagrams_dict, plugin_environment
):
    mock_requests_fallback(len(diagrams_dict))

    plant_uml_plugin.storage.data = diagrams_dict
    plant_uml_plugin.on_env(plugin_environment)

    for _, diagram in plant_uml_plugin.storage.items():
        assert isinstance(diagram.diagram, Fallback)

    assert len(plant_uml_plugin.storage.invalid) == len(diagrams_dict)


def test_on_post_page(plant_uml_plugin, diagrams_dict, html_page):
    plant_uml_plugin.storage.data = diagrams_dict
    output = plant_uml_plugin.on_post_page(html_page.content, html_page)

    assert output.count('<div class="puml light" style="">') == len(
        [True for v in diagrams_dict.values() if v.mode == ThemeMode.LIGHT]
    )
    assert output.count('<div class="puml dark" style="">') == len(
        [True for v in diagrams_dict.values() if v.mode == ThemeMode.DARK]
    )


def test_on_post_build(tmp_path, plant_uml_plugin):
    # Test if static files are correctly copied during the build process
    config = {"site_dir": str(tmp_path)}
    dest_dir = tmp_path.joinpath("assets/mkdocs_puml")
    os.makedirs(dest_dir)

    plant_uml_plugin.on_post_build(config)

    assert dest_dir.joinpath("puml.css").exists()
    assert dest_dir.joinpath("puml.js").exists()

    assert dest_dir.joinpath("interaction.css").exists()
    assert dest_dir.joinpath("interaction.js").exists()


def test_on_post_build_with_subdirectory(tmp_path, plant_uml_plugin):
    # Test if the plugin correctly handles subdirectories in the static folder
    config = {"site_dir": str(tmp_path)}

    plant_uml_plugin.on_post_build(config)

    dest_dir = tmp_path.joinpath("assets/mkdocs_puml")
    assert dest_dir.joinpath("puml.css").exists()


@pytest.mark.parametrize(
    "fallback,count,expected",
    [
        (
            0,
            Count(1, 0),
            "[dim][bold magenta]mkdocs_puml[/bold magenta]: Built 1 light diagram[/dim] [green bold]✔️[/green bold]",
        ),
        (
            0,
            Count(0, 1),
            "[dim][bold magenta]mkdocs_puml[/bold magenta]: Built 1 dark diagram[/dim] [green bold]✔️[/green bold]",
        ),
        (
            0,
            Count(2, 0),
            "[dim][bold magenta]mkdocs_puml[/bold magenta]: Built 2 light diagrams[/dim] [green bold]✔️[/green bold]",
        ),
        (
            0,
            Count(0, 2),
            "[dim][bold magenta]mkdocs_puml[/bold magenta]: Built 2 dark diagrams[/dim] [green bold]✔️[/green bold]",
        ),
        (
            0,
            Count(1, 1),
            "[dim][bold magenta]mkdocs_puml[/bold magenta]: Built 1 light and 1 dark diagrams[/dim] "
            "[green bold]✔️[/green bold]",
        ),
        (
            0,
            Count(0, 0),
            "[dim][bold magenta]mkdocs_puml[/bold magenta]: All diagrams loaded from cache"
            "[/dim] [green bold]✔️[/green bold]",
        ),
        (
            3,
            Count(3, 0),
            "[dim][bold magenta]mkdocs_puml[/bold magenta]: Built 3 light diagrams."
            "[/dim][bold red] 3 diagram failed to render ⨯[/bold red]",
        ),
    ],
)
def test_status_message(plant_uml_plugin, fallback, count, expected):
    msg = plant_uml_plugin._prepare_status_message(fallback, count)
    assert msg == expected



import pytest
from mkdocs_puml.model import Diagram, ThemeMode
from mkdocs_puml.storage import RAMStorage
from tests.plugins.conftest import is_uuid_valid


@pytest.fixture
def diagram_object():
    return Diagram("test", ThemeMode.LIGHT, "test")


def test_hash_light(diagram_object):
    storage = RAMStorage()
    assert is_uuid_valid(storage.hash(diagram_object))


def test_hash_dark(diagram_object):
    diagram_object.mode = ThemeMode.DARK
    storage = RAMStorage()

    h, _, dark = storage.hash(diagram_object).rpartition("-")
    assert is_uuid_valid(h)
    assert dark == "dark"


def test_add(diagram_object):
    storage = RAMStorage()
    key = storage.add(diagram_object)

    assert is_uuid_valid(key)
    assert len(storage.items()) == 1

    assert storage.items()[0][1] == diagram_object



import asyncio
from dataclasses import dataclass
import logging
import typing
import re

from urllib.parse import urljoin
from xml.dom.minidom import Element, parseString  # nosec

from httpx import AsyncClient, Response

from mkdocs_puml.encoder import encode
from mkdocs_puml.utils import sanitize_url


logger = logging.getLogger("mkdocs.plugins.plantuml")


@dataclass
class Fallback:
    """Fallback message for a scheme when PlantUML returns an error"""

    status_code: int
    message: str

    def __str__(self):  # pragma: no cover
        return f"{self.status_code}. {self.message}"


class PlantUML:
    """PlantUML converter class.
    It requests PUML service, updates received `svg`
    and returns it to the user.

    Attributes:
        base_url (str): Base URL to the PUML service
        num_workers (int): The size of pool to run requests in
        verify_ssl (bool): Designates whether the ``requests`` should verify SSL certiticate
        output_format (str): The output format for the diagrams (e.g., "svg" or "dsvg")

    Examples:
        Use this class as::

            puml = PlantUML("https://www.plantuml.com")
            svg = puml.translate([diagram])[0]
    """

    _html_comment_regex = re.compile(r"<!--.*?-->", flags=re.DOTALL)

    def __init__(
        self,
        base_url: str,
        verify_ssl: bool = True,
        output_format: str = "svg",
        timeout: int = 40,
    ):
        # Use sanitize_url because urllib removes last part of url which doesn't
        # end with / which makes it inconvenient to work with.
        self.base_url = sanitize_url(base_url)
        self.base_url = f"{self.base_url}{output_format}/"

        self.verify_ssl = verify_ssl
        self.timeout = timeout

    def translate(self, schemes: typing.Iterable[str]) -> typing.List[typing.Union[str, Fallback]]:
        """Translate PlantUML schemes into the received SVG image.

        Args:
            schemes (list): string representation of PUML diagram

        Returns:
            SVG image of built diagram
        """
        encoded = [self.preprocess(v) for v in schemes]

        svg_images = self.request(encoded)

        for i, v in enumerate(svg_images):
            svg_images[i] = self.postprocess(v)

        return svg_images

    def preprocess(self, content: str) -> str:
        """Pre-process the content before passing it
        to the PlantUML service.

        Encoding of the content should be
        done in the step of preprocessing.

        Args:
            content (str): string representation PUML diagram
        Returns:
            Encoded and pre-processed PUML diagram
        """
        return encode(content)

    def postprocess(self, content: typing.Union[str, Fallback]) -> typing.Union[Fallback]:
        """Postprocess an SVG diagram received from PlantUML server.

        The code that applies CSS styling to the SVG can be placed here.

        Args:
            content (str): SVG representation of build diagram
        Returns:
            Post-processed SVG diagram
        """
        if isinstance(content, Fallback):
            return content

        diagram_content = self._clean_comments(content)

        svg = self._convert_to_dom(diagram_content)
        self._stylize_svg(svg)

        return svg.toxml()

    def request(self, schemes: list[str]) -> list[typing.Union[str, Fallback]]:
        """Request PlantUML service with the encoded diagram;
        return SVG content

        Args:
            schemes (str): Encoded string representation of the diagram diagrams
        Returns:
            SVG representation of the diagram
        """
        responses: list[Response] = asyncio.run(self._request_all(schemes))

        svgs = []
        for s, resp in zip(schemes, responses):
            # Use 'ignore' to strip non-utf chars
            c = resp.content.decode("utf-8", errors="ignore")
            if not resp.is_success:
                logger.warning(
                    f"While building diagram \n\n{s}\n\nServer responded"
                    f" with a status {resp.status_code}"
                )
                svgs.append(Fallback(status_code=resp.status_code, message=c))
            else:
                svgs.append(c)

        return svgs

    async def _request_one(self, uri: str) -> Response:
        """Request request PlantUML server asynchronously

        Args:
            uri (str): URI with encoded diagram attached to it

        Returns:
            Response: response from PlantUML server
        """
        async with AsyncClient(verify=self.verify_ssl, timeout=self.timeout) as client:
            return await client.get(uri)

    async def _request_all(self, schemes: list[str]):
        """Asynchronous wrapper that creates request coroutine for
        each scheme and after await returns an ordered list of responses.

        Args:
            schemes (list[str]): encoded PlantUML diagrams

        Returns:
            list[Response]: ordered list of Responses
        """
        return await asyncio.gather(
            *(self._request_one(urljoin(self.base_url, v)) for v in schemes)
        )

    def _clean_comments(self, content: str) -> str:
        """Remove comments from HTML content"""
        return self._html_comment_regex.sub("", content)

    def _convert_to_dom(self, content: str) -> Element:
        """The method to convert received SVG into XML DOM
        for future modifications
        """
        dom = parseString(content)  # nosec
        svg = dom.getElementsByTagName("svg")[0]
        svg.setAttribute("class", "diagram")
        return svg

    def _stylize_svg(self, svg: Element):
        """This method is used for SVG tags modifications"""
        svg.setAttribute("preserveAspectRatio", "xMidYMid meet")



from enum import Enum
from mkdocs.config.base import Config
from mkdocs.config.config_options import Type, SubConfig, Choice


class CacheBackend(Enum):
    DISABLED = "disabled"
    LOCAL = "local"

    @classmethod
    def values(cls):
        return [v.value for v in cls]


class ThemeConfig(Config):
    enabled = Type(bool, default=True)
    light = Type(str, default="default/light")
    dark = Type(str, default="default/dark")

    url = Type(
        str,
        default="https://raw.githubusercontent.com/MikhailKravets/mkdocs_puml/master/themes/",
    )


class LocalCacheConfig(Config):
    path = Type(str, default="~/.cache/mkdocs_puml/")


class CacheConfig(Config):
    backend = Choice(CacheBackend.values(), default=CacheBackend.LOCAL.value)
    local = SubConfig(LocalCacheConfig)


class InteractionConfig(Config):
    enabled = Type(bool, default=True)


class PlantUMLConfig(Config):
    puml_url = Type(str)
    puml_keyword = Type(str, default="puml")
    verify_ssl = Type(bool, default=True)
    verbose = Type(bool, default=True)
    theme = SubConfig(ThemeConfig)  # SubConfig already has an `{}` as default
    cache = SubConfig(CacheConfig)
    interaction = SubConfig(InteractionConfig)



import dataclasses
import hashlib
from unittest.mock import mock_open

import msgpack
import pytest
from mkdocs_puml.model import Diagram, ThemeMode
from mkdocs_puml.puml import Fallback


@pytest.fixture
def hash_diagrams():
    d = [
        Diagram("test_one", ThemeMode.LIGHT, "test"),
        Diagram("test_two", ThemeMode.DARK, "test"),
    ]

    return {hashlib.blake2b(v.scheme.encode("utf-8")).hexdigest(): v for v in d}


@pytest.fixture
def hash_diagrams_with_fallback():
    d = [
        Diagram("test_one", ThemeMode.LIGHT, Fallback(400, "Test")),
        Diagram("test_two", ThemeMode.DARK, "test"),
    ]

    return {hashlib.blake2b(v.scheme.encode("utf-8")).hexdigest(): v for v in d}


@pytest.fixture
def patch_stream(monkeypatch, hash_diagrams):
    encoded = msgpack.dumps(
        {k: dataclasses.asdict(v) for k, v in hash_diagrams.items()}
    )
    mock = mock_open(read_data=encoded)
    monkeypatch.setattr("builtins.open", mock)
    return mock



from pathlib import Path
from unittest.mock import MagicMock

import pytest

from mkdocs_puml.model import Diagram, ThemeMode
from mkdocs_puml.storage import FileStorage


# TODO: find a better way to mock Path.
@pytest.fixture
def patch_exist_stat(monkeypatch, patch_path_mkdir):
    stat = MagicMock()
    stat.return_value = MagicMock()
    stat.return_value.st_size = 2

    exists = MagicMock()
    exists.return_value = True
    monkeypatch.setattr("pathlib.Path.exists", exists)
    monkeypatch.setattr("pathlib.Path.stat", stat)


def test_read_nonexistent_file(monkeypatch, patch_path_mkdir):
    exists = MagicMock()
    exists.return_value = False

    monkeypatch.setattr("pathlib.Path.exists", exists)

    fs = FileStorage(Path("test"), "test.mock")

    assert exists.call_count == 1
    assert len(fs.data) == 0
    assert len(fs.invalid) == 0


def test_read_empty_file(monkeypatch, patch_path_mkdir):
    stat = MagicMock()
    stat.return_value = MagicMock()
    stat.return_value.st_size = 0

    exists = MagicMock()
    exists.return_value = True
    monkeypatch.setattr("pathlib.Path.exists", exists)
    monkeypatch.setattr("pathlib.Path.stat", stat)

    fs = FileStorage(Path("test"), "test.mock")

    assert stat.call_count == 1
    assert len(fs.data) == 0
    assert len(fs.invalid) == 0


def test_read(monkeypatch, patch_exist_stat, patch_stream, hash_diagrams):
    fs = FileStorage(Path("test"), "test.mock")

    assert len(fs.data) == len(fs.invalid) == len(hash_diagrams)

    for k, v in hash_diagrams.items():
        assert fs[k] == v
        assert k in fs.invalid


def test_add_new(patch_exist_stat, patch_stream, hash_diagrams):
    fs = FileStorage(Path("test"), "test.mock")
    d = Diagram("test_new", ThemeMode.DARK, "test")
    fs.add(d)

    assert len(fs.data) == len(hash_diagrams) + 1
    assert len(fs.invalid) == len(hash_diagrams)


def test_count_total(patch_exist_stat, patch_stream, hash_diagrams):
    fs = FileStorage(Path("test"), "test.mock")
    d = Diagram("test_new", ThemeMode.DARK, "test")
    fs.add(d)

    c = fs.count(True)
    assert c.light == 1
    assert c.dark == 2


def test_count_to_request(patch_exist_stat, patch_stream, hash_diagrams):
    fs = FileStorage(Path("test"), "test.mock")
    d = Diagram("test_new", ThemeMode.DARK, None)
    fs.add(d)

    c = fs.count()
    assert c.light == 0
    assert c.dark == 1


def test_add_existing(patch_exist_stat, patch_stream, hash_diagrams):
    fs = FileStorage(Path("test"), "test.mock")

    diagram_list = list(hash_diagrams.items())
    fs.add(diagram_list[0][1])

    assert len(fs.data) == len(hash_diagrams)
    assert len(fs.invalid) == len(hash_diagrams) - 1

    assert diagram_list[0][0] not in fs.invalid


def test_update(patch_exist_stat, patch_stream, hash_diagrams_with_fallback):
    fs = FileStorage(Path("test"), "test.mock")
    fs.update(hash_diagrams_with_fallback.items())

    # One from patch_stream, and one Fallback from hash_diagrams_with_fallback
    assert len(fs.invalid) == 1 + 1


def test_save(patch_exist_stat, patch_stream, hash_diagrams):
    fs = FileStorage(Path("test"), "test.mock")

    fs.invalid.remove(list(hash_diagrams.items())[0][0])
    fs.save()

    # One from patch_stream, and one Fallback from hash_diagrams_with_fallback
    assert patch_stream.return_value.write.call_count == 1



from pathlib import Path
import typing
import re
import os
import shutil

from rich.console import Console

from mkdocs.config.defaults import MkDocsConfig
from mkdocs.plugins import BasePlugin

from mkdocs_puml.config import PlantUMLConfig
from mkdocs_puml.model import Count, Diagram, ThemeMode
from mkdocs_puml.storage import AbstractStorage, build_storage
from mkdocs_puml.puml import Fallback, PlantUML
from mkdocs_puml.theme import Theme


class PlantUMLPlugin(BasePlugin[PlantUMLConfig]):
    """MKDocs plugin that converts puml diagrams into SVG images.

    It works only with a remote PlantUML service. You should add
    these configs into ``mkdocs.yml``::

            plugins:
                - mkdocs_puml:
                    puml_url: https://www.plantuml.com/plantuml

    The rest of configuration is optional. Please refer to plugin
    documentation to view them all

    Attributes:
        pre_class_name (str): the class that will be set to intermediate <pre> tag
                              containing uuid code
        container (str): html element where the diagrams will be inserted.
                         **DO NOT** insert any `\n` characters, as the Markdown parser will convert them into
                         `<p>...</p>`, which may result in an unexpected html
    """

    pre_class_name = "diagram-key"
    container = "<div class='puml-container'>{}</div>"

    def __init__(self):
        self.regex: typing.Optional[typing.Any] = None
        self.uuid_regex = re.compile(
            rf'<pre class="{self.pre_class_name}">(.+?)</pre>', flags=re.DOTALL
        )

        self.puml: typing.Optional[PlantUML] = None
        self.themer: typing.Optional[Theme] = None
        self.storage: typing.Optional[AbstractStorage] = None
        self.console: typing.Optional[Console] = None

    def on_config(self, config: MkDocsConfig) -> MkDocsConfig:
        """Event that is fired by mkdocs when configs are created.

        All required classes such as PlantUML, Theme, or any class for storage
        are initialized in this method.
        Also, `puml.css` file that enable dark / light mode styles is added to `extra_css`.

        Args:
            config: Full mkdocs.yml config file. To access configs of PlantUMLPlugin only,
                    use self.config attribute.

        Returns:
            Full config of the mkdocs
        """
        config["extra_css"].append("assets/mkdocs_puml/puml.css")
        config["extra_javascript"].append("assets/mkdocs_puml/puml.js")

        if self.config.interaction.enabled:
            config["extra_css"].append("assets/mkdocs_puml/interaction.css")
            config["extra_javascript"].extend(
                [
                    "https://unpkg.com/@panzoom/panzoom@4.5.1/dist/panzoom.min.js",
                    "assets/mkdocs_puml/interaction.js",
                ]
            )

        self.console = Console(quiet=not self.config.verbose)
        self.puml = PlantUML(
            self.config.puml_url,
            verify_ssl=self.config.verify_ssl,
        )
        self.puml_keyword = self.config.puml_keyword
        self.regex = re.compile(rf"```{self.puml_keyword}(\n.+?)```", flags=re.DOTALL)

        if self.config.theme.enabled:
            self.themer = Theme(self.config.theme.url)

            self.theme_light = self.config.theme.light
            self.theme_dark = self.config.theme.dark
        else:
            self.themer = None
            self.theme_light = None
            self.theme_dark = None

        self.storage = build_storage(self.config.cache)

        return config

    def on_page_markdown(self, markdown: str, *args, **kwargs) -> str:
        """Event to fire for each .md page.

        Here, all ``puml`` code blocks are found and added to a storage.

        Then, <pre class="...">{key of diagram}</pre> tags are added to
        the markdown page.

        Args:
            markdown: Markdown page in which to look for PlantUML diagrams.

        Returns:
            Updated markdown page
        """
        with self.console.status(
            "[bold dim cyan]Search puml in markdown",
            spinner="dots2",
            spinner_style="magenta",
        ):
            schemes = self.regex.findall(markdown)

            # DO NOT insert `\n` characters in the replacement!
            for v in schemes:
                if self.themer:
                    replace_into = self.container.format(self._store_dual(v))
                else:
                    replace_into = self.container.format(self._store_single(v))
                markdown = markdown.replace(
                    f"```{self.puml_keyword}{v}```",
                    replace_into,
                )

        return markdown

    def _store_single(self, scheme: str) -> str:
        d = Diagram(scheme, mode=ThemeMode.LIGHT)
        key = self.storage.add(d)
        return f'<pre class="{self.pre_class_name}">{key}</pre>'

    def _store_dual(self, scheme: str) -> str:
        d_light = Diagram(
            self.themer.include(self.config.theme.light, scheme), mode=ThemeMode.LIGHT
        )
        d_dark = Diagram(
            self.themer.include(self.config.theme.dark, scheme), mode=ThemeMode.DARK
        )

        key_light = self.storage.add(d_light)
        key_dark = self.storage.add(d_dark)

        return (
            f'<pre class="{self.pre_class_name}">{key_light}</pre>'
            f'<pre class="{self.pre_class_name}">{key_dark}</pre>'
        )

    def on_env(self, env, *args, **kwargs):
        """The event is fired when jinja environment is configured.
        Such as it is fired once when all .md pages are processed,
        we can use it to request PlantUML service and convert the
        diagrams.

        Args:
            env: jinja environment
        Returns:
            Jinja environment
        """
        with self.console.status(
            "[bold dim cyan]Building PlantUML diagrams",
            spinner="dots2",
            spinner_style="magenta",
        ):
            to_request = self.storage.schemes()
            to_req_count = self.storage.count()
            svgs = self.puml.translate(to_request.values())
            self.storage.update(zip(to_request.keys(), svgs))

            fallback_count = len([True for v in svgs if isinstance(v, Fallback)])

        self.console.print(self._prepare_status_message(fallback_count, to_req_count))
        return env

    def on_post_page(self, output: str, page, *args, **kwargs) -> str:
        """The event is fired after HTML page is rendered.
        Here, we substitute <pre> tags with the corresponding SVG images.

        Args:
            output: rendered HTML in str format
            page: Page object

        Returns:
            HTML page containing SVG diagrams
        """
        schemes = self.uuid_regex.findall(output)
        for v in schemes:
            output = self._replace(v, output)
            page.content = output

        return output

    def on_post_build(self, config):
        """Event triggered after the build process is complete.

        This method copies static assets of the plugin and saves
        the diagrams to the storage.

        Args:
            config (dict): The MkDocs configuration object.

        """
        # Path to the static directory in the plugin
        static_dir = Path(__file__).parent.joinpath("static")
        # Destination directory in the site output
        dest_dir = Path(config["site_dir"]).joinpath("assets/mkdocs_puml/")

        if not dest_dir.exists():
            os.makedirs(dest_dir)

        # shutil.copy(static_dir, dest_dir)
        # shutil.copy(puml_js, dest_dir)
        shutil.copytree(static_dir, dest_dir, dirs_exist_ok=True)

        self.storage.save()

    def _replace(self, key: str, content: str) -> str:
        """Replace a key of a diagram with a diagram svg in a
        content
        """
        diagram = self.storage[key]

        # When theming is not enabled, user will manually manage themes in each diagram.
        # Also, only one version of diagram will be generated for each scheme, which
        # should be displayed always despite the light / dark mode of mkdocs-material.
        style = "display: block" if not self.config.theme.enabled else ""
        replacement = (
            f'<div class="puml {diagram.mode}" style="{style}">{diagram.diagram}</div>'
        )
        return content.replace(
            f'<pre class="{self.pre_class_name}">{key}</pre>', replacement
        )

    def _prepare_status_message(self, fallback_count: int, req_count: Count):
        if fallback_count:
            ok_msg = f".[/dim][bold red] {fallback_count} diagram failed to render ⨯[/bold red]"
        else:
            ok_msg = "[/dim] [green bold]✔️[/green bold]"

        if req_count.light == 0 and req_count.dark == 0:
            built_msg = "All diagrams loaded from cache"
        elif req_count.light == 0:
            d = "diagram" if req_count.dark == 1 else "diagrams"
            built_msg = f"Built {req_count.dark} dark {d}"
        elif req_count.dark == 0:
            d = "diagram" if req_count.light == 1 else "diagrams"
            built_msg = f"Built {req_count.light} light {d}"
        else:
            built_msg = f"Built {req_count.light} light and {req_count.dark} dark diagrams"

        return "[dim][bold magenta]mkdocs_puml[/bold magenta]: " + built_msg + ok_msg



def sanitize_url(url: str) -> str:
    """Converts a url to a normalized state.

    ## Validators

    1. Ensures that url ends with /

    Args:
        url (str): url to sanitize

    Returns:
        str: sanitized url
    """
    return url if url.endswith("/") else f"{url}/"



from dataclasses import dataclass
from typing import Optional
import typing

from mkdocs_puml.puml import Fallback


class ThemeMode:
    LIGHT = "light"
    DARK = "dark"


@dataclass
class Diagram:
    scheme: str
    mode: ThemeMode
    diagram: Optional[typing.Union[str, Fallback]] = None


@dataclass
class Count:
    light: int
    dark: int



import re
from mkdocs_puml.utils import sanitize_url

C4_REGEX = re.compile(r"(!include(?:.+)(?:[Cc]4)(?:.+).puml)")


class Theme:
    """Theme class helps integrate available themes into PlantUML code.

    Theme includes theme url after all C4 inclusions. It makes possible
    to provide custom styling to C4 code as well.

    Args:
        url (str): repository of themes
    """

    def __init__(self, url: str):
        self.url = sanitize_url(url)

    def include(self, theme: str, diagram: str) -> str:
        """Includes theme to the beginning of PlantUML diagram

        Args:
            theme (str): theme name to include
            diagram (str): diagram into which to include

        Returns:
            str: PlantUML diagram with theme included
        """
        diagram = diagram.strip()
        with_c4 = C4_REGEX.split(diagram)
        url = self.url_for(theme)

        if len(with_c4) == 1:
            if diagram.startswith("@startuml"):
                head, _, tail = diagram.partition("\n")
            else:
                head, tail = None, diagram

            if head:
                return f"{head}\n!include {url}\n{tail}"
            return f"!include {url}\n{tail}"
        else:
            tail = with_c4[-1]
            with_c4[-1] = f"\n!include {url}"
            with_c4.append(tail)
            return "".join(with_c4)

    def url_for(self, theme: str) -> str:
        """Create full url for theme

        Args:
            theme (str): theme name

        Returns:
            str: full url for theme
        """
        return f"{self.url}{theme}.puml"



from abc import ABC, abstractmethod
import dataclasses
import hashlib
from pathlib import Path
from typing import Iterable
import typing
import uuid

import msgpack

from mkdocs_puml.config import CacheBackend, CacheConfig
from mkdocs_puml.model import Count, Diagram, ThemeMode
from mkdocs_puml.puml import Fallback


class AbstractStorage(ABC):
    """PlantUML may take up to several seconds to render
    a single diagram. Storage adds a persistence to the built SVG,
    allowing to use it as a cache for the diagrams.
    """

    def __init__(self):
        self.data: dict[str, Diagram] = {}

        # this is a set of keys that should not be saved
        # in the next save(..) iteration
        self.invalid: set[str] = set()

    @abstractmethod
    def add(self, d: Diagram) -> str:  # pragma: no cover
        """Add a diagram to the storage and return
        a key of the diagram.

        Args:
            d (Diagram): diagram object to add to the storage

        Returns:
            str: key of the diagram
        """

    def update(self, svg: Iterable[tuple[str, typing.Union[str, Fallback]]]):
        """Update a collection of diagrams from an
        iterable of SVG images.

        Args:
            svg (Iterable[tuple[str, str | Fallback]]): iterable of tuples `(diagram key, svg)`
        """
        for key, s in svg:
            if isinstance(s, Fallback):
                self.invalid.add(key)
            self.data[key].diagram = s

    @abstractmethod
    def hash(self, d: Diagram) -> str:  # pragma: no cover
        """Returns a collision-free hash for the diagram

        Args:
            d (Diagram): diagram object

        Returns:
            str: string hash of the diagram
        """

    @abstractmethod
    def save(self):  # pragma: no cover
        """Saves data to a persistent storage: a file,
        database, etc
        """

    def schemes(self) -> dict[str, str]:
        """A dictionary of diagrams that doesn't have SVG
        image rendered for them.

        Returns:
            dict[str, str]: dictionary where key is diagram key
                            and value is diagram scheme
        """
        return {k: v.scheme for k, v in self.data.items() if v.diagram is None}

    def items(self) -> list[str, Diagram]:
        """A list of (key, Dictionary) tuples"""
        return [(k, v) for k, v in self.data.items()]

    def keys(self) -> Iterable[str]:
        """Iterable of diagram keys"""
        return self.data.keys()

    def count(self, total: bool = False) -> Count:
        light, dark = 0, 0
        for v in self.data.values():
            if total:
                if v.mode == ThemeMode.LIGHT:
                    light += 1
                else:
                    dark += 1
            elif v.diagram is None:
                if v.mode == ThemeMode.LIGHT:
                    light += 1
                else:
                    dark += 1

        return Count(light, dark)

    def __getitem__(self, key: str) -> Diagram:
        """Get diagram by key"""
        return self.data[key]


class RAMStorage(AbstractStorage):
    """RAMStorage doesn't provide persistence.
    The diagrams are stored in memory only.

    It may be useful when user wants to disable caching.
    """

    def hash(self, d: Diagram):
        if d.mode == ThemeMode.LIGHT:
            return str(uuid.uuid4())
        elif d.mode == ThemeMode.DARK:
            return f"{uuid.uuid4()}-dark"

    def add(self, d: Diagram):
        h = self.hash(d)
        self.data[h] = d
        return h

    def save(self):  # pragma: no coverage
        """RAMStorage keeps diagrams in RAM only"""


class FileStorage(AbstractStorage):
    """`FileStorage` handles diagrams stored in a file
    using the MessagePack format.

    This class can be used as a cache, sending only diagrams
    without an SVG image to the PlantUML server.

    `FileStorage` uses `blake2b` as a hasher for diagrams.

    Args:
        base_dir (Path): the directory where `FileStorage` stores the file.
        filename (str, optional): name of the file. Defaults to "storage.mpack".
    """

    def __init__(self, base_dir: Path, filename: str = "storage.mpack"):
        super().__init__()

        work_dir = Path.cwd().name
        dir = base_dir.expanduser().joinpath(work_dir)
        dir.mkdir(parents=True, exist_ok=True)

        self.path = dir.joinpath(filename)
        self._read_data()

        # This attribute helps in finding invalid diagrams
        # that don't exists in docs anymore but present in storage.
        #
        # During the build self.data will be populated by call of
        # self.add method. If the key was never added through self.add
        # method, then this diagram doesn't exist anymore in the docs
        # and it can be safely deleted from storage.
        self.invalid = set(self.data.keys())

    def hash(self, d: Diagram):
        return hashlib.blake2b(d.scheme.encode("utf-8")).hexdigest()

    def add(self, d: Diagram):
        h = self.hash(d)

        if h in self.invalid:
            self.invalid.remove(h)

        if h not in self.data:
            self.data[h] = d

        return h

    def save(self):
        with open(self.path, "wb") as f:
            to_save = {}
            for k, v in self.data.items():
                if k not in self.invalid:
                    to_save[k] = dataclasses.asdict(v)
            msgpack.dump(to_save, f)

    def _read_data(self):
        if not self.path.exists() or self.path.stat().st_size == 0:
            return

        with open(self.path, "rb") as f:
            raw = msgpack.load(f)
            self.data = {k: Diagram(**v) for k, v in raw.items()}


def build_storage(config: CacheConfig) -> AbstractStorage:
    """Factory function that returns a storage class instance
    based on the `CacheConfig`.

    Cache backend:

    * `disabled` — build `RAMStorage` instance
    * `local` — build `FileStorage` instance
    """
    if config.backend == CacheBackend.DISABLED.value:
        return RAMStorage()
    elif config.backend == CacheBackend.LOCAL.value:
        return FileStorage(Path(config.local.path))
